import logging

from lytix_py.LAsyncStore.LAsyncStore import LAsyncStore, LAsyncStoreClass

class LLoggerStreamWrapper(logging.StreamHandler):
    customAsyncStore: LAsyncStoreClass  = None
    def __init__(self, customAsyncStore: LAsyncStoreClass = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if customAsyncStore:
            self.customAsyncStore = customAsyncStore

    def emit(self, record):
        try:
            formatted = self.format(record)
            if self.customAsyncStore:
                self.customAsyncStore.appendToLogs(formatted)
            else:
                LAsyncStore.appendToLogs(formatted)
        except Exception as e:
            pass






import asyncio
import multiprocessing
import os
import sys
import threading
import time
import uuid
from lytix_py.Lytix.LytixTest import lytixTest
from lytix_py.MetricCollector.MetricCollector import MetricCollector
from lytix_py.colors import colors


def printTestRunning():
    while True:
        print(f"{colors.YELLOW}🐶 Waiting for Lytix Response...{colors.RESET}")
        time.sleep(2.5)


async def runTestSuite():
    """
    Gather all @lytix.test decorators present and
    run the test suite
    """
    allTests = lytixTest._testState

    print(
        f"{colors.GREEN}Starting test suite for: {len(allTests)} test(s)...{colors.RESET}"
    )

    """
    First create a test suite
    """
    testSuiteId = MetricCollector._createTestSuite()

    """
    Then all each function
    """
    asyncFunctions = []
    syncFunctions = []
    for func in allTests.values():
        if asyncio.iscoroutinefunction(func["function"]):
            asyncFunctions.append(func)
        else:
            syncFunctions.append(func)

    """
    Keep track of which ones passed and failed so we can report at the end"""
    testsPassed = {}
    testsFailed = {}
    totalTests = len(asyncFunctions) + len(syncFunctions)

    # testRunning = True

    printThread = multiprocessing.Process(target=printTestRunning, args=())
    printThread.start()

    # bLoop = asyncio.new_event_loop()
    # bLoop.run_until_complete(printTestRunning())

    # backgroundPrint = asyncio.create_task(printTestRunning())
    # backgroundPrint = loop.run_coroutine_threadsafe(printTestRunning(), loop)

    # printTask = asyncio.ensure_future(printTestRunning())

    """
    Gather and run the async tests if possible
    """
    for func in asyncFunctions:
        functionName = func["function"].__name__

        @lytixTest.test(
            **func["testArgs"], _testSuiteId=testSuiteId, _testName=functionName
        )
        async def wrapper(*args, **kwargs):
            results = await func["function"](*args, **kwargs)
            return results

        testResult = await wrapper()
        if testResult == "FAILED":
            testsFailed[functionName] = testResult
            print(f"{colors.RED}Test failed for {functionName}{colors.RESET}")
        elif testResult == "SUCCESS":
            testsPassed[functionName] = testResult
            print(f"{colors.GREEN}Test passed for {functionName}{colors.RESET}")
        else:
            print(
                f"{colors.YELLOW}Test result is {testResult} for {functionName} which is unhandled{colors.RESET}"
            )

    """
    Then run all the sync tests one by one
    """
    for func in syncFunctions:
        functionName = func["function"].__name__

        @lytixTest.test(
            **func["testArgs"], _testSuiteId=testSuiteId, _testName=functionName
        )
        def wrapper(*args, **kwargs):
            return func["function"](*args, **kwargs)

        testResult = wrapper()
        if testResult == "FAILED":
            testsFailed[functionName] = testResult
            print(f"{colors.RED}Test failed for {functionName}{colors.RESET}")
        elif testResult == "SUCCESS":
            testsPassed[functionName] = testResult
            print(f"{colors.GREEN}Test passed for {functionName}{colors.RESET}")
        else:
            print(
                f"{colors.YELLOW}Test result is {testResult} for {functionName} which is unhandled{colors.RESET}"
            )

    """
    Report the final results
    """
    printThread.terminate()
    finalizeRes = MetricCollector._finalizeTestGroup(testSuiteId)
    if finalizeRes["valid"] is not True:
        print(
            f"{colors.RED}Failed to finalize test group {testSuiteId} with error: {finalizeRes['error']}{colors.RESET}"
        )
    # testRunning = False
    # backgroundPrint.cancel()
    if len(testsFailed) == 0:
        # All passed
        print(
            f"{colors.GREEN}{len(testsPassed)}/{totalTests} tests passed! 🍾{colors.RESET}"
        )
    else:
        # Some failed
        print(
            f"{colors.RED}{len(testsPassed)}/{totalTests} tests passed!{colors.RESET}"
        )

    for test in testsPassed:
        print(f"{colors.GREEN}-> {test}: passed ✅{colors.RESET}")
    for test in testsFailed:
        print(f"{colors.RED}-> {test}: failed ❌{colors.RESET}")

    if len(testsFailed) > 0:
        sys.exit(1)



from datetime import datetime
import json

from lytix_py.LAsyncStore.LAsyncStore import LAsyncStore
from lytix_py.MetricCollector.MetricCollector import MetricCollector


def LErrorIncrement(errorMsg: str, errorMetadata: dict = {}):
    try:
        """
        First get all logs from our async store
        """
        logs = LAsyncStore.getLogs()

        """
        Add a new log with our error
        """
        logs.append(
            json.dumps(
                {
                    "name": "LError",
                    "hostname": "",
                    "pid": -1,
                    "level": 50,
                    "msg": errorMsg,
                    "time": datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
                }
            )
        )

        """
        Send the logs+ metadata to lytix
        """
        MetricCollector._captureLError(
            errorMsg=errorMsg, errorMetadata=errorMetadata, logs=logs
        )
    except Exception as e:
        print("Error sending Lytix metric", e)



import asyncio
import json
import logging
import socket
import sys
from typing import Mapping

from lytix_py.LAsyncStore.LAsyncStore import LAsyncStore, LAsyncStoreClass
from lytix_py.LLogger.LLoggerStreamWrapper import LLoggerStreamWrapper


class LLogger:
    metadata: dict = None

    """
    @param config: Optional dict of the following shape: { console: boolean }
    """

    def __init__(
        self,
        loggerName: str,
        metadata: dict = None,
        asyncStore: LAsyncStoreClass = None,
    ):
        self.logger = logging.getLogger(loggerName)
        if not len(self.logger.handlers):
            hostname = socket.gethostname()

            """
            Setup formatting and logging config
            """
            self.logger.setLevel("INFO")
            fmt = (
                '{"time": "%(asctime)s.%(msecs)03d",  "hostname": "'
                + hostname
                + '",  "level": "%(levelname)s", "msg": "%(message)s",  "name": "%(name)s", "pid": "%(process)d"}'
            )

            datefmt = "%Y-%m-%d,%H:%M:%S"
            formatter = logging.Formatter(fmt, datefmt)
            handler = logging.StreamHandler(stream=sys.stdout)
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)

            progress = LLoggerStreamWrapper(customAsyncStore=asyncStore)
            progress.setFormatter(formatter)
            self.logger.addHandler(progress)

        """
        Define our async store to store recent logs
        """
        if asyncStore:
            self.asyncStore = asyncStore
        else:
            self.asyncStore = LAsyncStore

        if metadata:
            self.metadata = metadata

    def info(self, msg: object):
        toLog = f"{self.getMetadataLoggerString()}{self.stringify(msg)}"
        self.logger.info(toLog)

    def warn(self, msg: object):
        toLog = f"{self.getMetadataLoggerString()}{self.stringify(msg)}"
        self.logger.warn(toLog)

    def error(self, msg: object):
        toLog = f"{self.getMetadataLoggerString()}{self.stringify(msg)}"
        self.logger.error(toLog)

    def debug(self, msg: object):
        toLog = f"{self.getMetadataLoggerString()}{self.stringify(msg)}"
        self.logger.debug(toLog)

    """"
    Stringify and remove leading and trailing quotes
    """

    def stringify(self, msg: object):
        return json.dumps(msg).strip('"')

    """
    Set metadata to be logged
    """

    def setMetadata(self, metadata: dict):
        self.metadata = metadata

    """
    Get a pretty printed version of the metadata stored to attach to a log
    """

    def getMetadataLoggerString(self):
        metadata = self.metadata
        if metadata and len(metadata) > 0:
            to_return = "["
            for key, value in metadata.items():
                to_return += f"{key}={value};"
            to_return = to_return[:-1]
            to_return += "] "
            return to_return
        return ""

    def runInAsyncContext(self, func):
        asyncio.run(func())



import asyncio
from collections import defaultdict
import contextvars
from functools import wraps
import logging
import time
from typing import Any, Callable, DefaultDict, List, TypeVar, cast
import uuid

from lytix_py.MetricCollector.MetricCollector import MetricCollector

"""
@see https://mypy.readthedocs.io/en/stable/generics.html#declaring-decorators
"""
F = TypeVar("F", bound=Callable[..., Any])


_lytixTestContext: contextvars.ContextVar[DefaultDict[str, str]] = (
    contextvars.ContextVar(
        "lytixTestContext",
        default=defaultdict(
            lambda: {
                "traceId": None,
                "modelName": None,
                "messages": None,
            },
        ),
    )
)


class LytixTestWrapper:
    _logger = logging.getLogger("lytix-test")
    _testState = {}

    def test(
        self,
        *kwargs,
        testsToRun: List[str],
        _testSuiteId: str = None,
        _testName: str = None,
    ) -> Callable[[F], F]:
        """
        Main testing decorator.
        @param testGroupId: This is set when the test suite runs
        @param testName: This is set when the test suite runs
        """

        def decorator(func: F) -> F:
            """
            If we have any function args, it can't be a test
            """
            if kwargs:
                self._logger.error("Lytix tests cannot have function args")
            else:
                self._testState[func.__name__] = {
                    "function": func,
                    "testArgs": {"testsToRun": testsToRun},
                }

            """
            First lets see if this is an async function
            """
            isAsyncFunction = asyncio.iscoroutinefunction(func)

            """
            Wrap our function in a context state
            """
            if isAsyncFunction:
                return self._asyncTestWrapper(func, testsToRun, _testSuiteId, _testName)
            else:
                return self._syncTestWrapper(func, testsToRun, _testSuiteId, _testName)

        return decorator

    def _syncTestWrapper(
        self, func: F, testsToRun: List[str], testSuiteId: str, testName: str
    ):
        """
        Sync wrapper for the decorator
        """

        @wraps(func)
        def syncDecorator(*args, **kwargs):
            testContext = self._prepTestDecoratorContext(testsToRun)
            func(*args, **kwargs)

            """
            Before returning to the user, send the test + IO back to HQ
            """
            testResult = self._startAndWaitForLytixTest(testSuiteId, testName)

            return testResult

        return cast(F, syncDecorator)

    def _asyncTestWrapper(
        self, func: F, testsToRun: List[str], testSuiteId: str, testName: str
    ):
        """
        Async wrapper for the decorator
        """

        @wraps(func)
        async def asyncDecorator(*args, **kwargs):
            testContext = self._prepTestDecoratorContext(testsToRun)
            await func(*args, **kwargs)

            """
            Before returning to the user, send the test + IO back to HQ
            """
            testResult = self._startAndWaitForLytixTest(testSuiteId, testName)

            return testResult

        return cast(F, asyncDecorator)

    def _startAndWaitForLytixTest(self, testSuiteId: str, testName: str):
        """
        Starts a new test in Lytix and waits for it to complete
        """
        context, valid = self._getTestContext()
        if not valid:
            return

        """
        Validate we have everything we need
        """
        messages = context["messages"]
        testsToRun = context["testsToRun"]
        traceId = context["traceId"]
        if messages is None or testsToRun is None or traceId is None:
            self._logger.error(
                "Missing input, output, testsToRun, or traceId in context, make sure you have setup the correct decorator"
            )
            return False

        """
        Start the test
        """
        testRunIds = MetricCollector._kickoffTestRun(
            messages, testsToRun, traceId, testSuiteId, testName
        )

        if testRunIds is False:
            return "FAILED"

        """
        Wait for it to finish
        """
        allFinished = False
        anyFailed = False
        while allFinished is False:
            allFinished = True
            """
            Loop over all testRunIds
            """
            for testRunId in testRunIds:
                testStatus = MetricCollector._getTestStatus(testRunId)
                if testStatus["testResult"] == "FAILED":
                    anyFailed = True
                if testStatus["testStatus"] not in ["FAILED", "SUCCESS"]:
                    allFinished = False
                    break

            if allFinished is False:
                time.sleep(5)

        """
        Result can be FAILED or SUCCESS, if it's FAILED
        """
        if anyFailed:
            return "FAILED"
        else:
            return "SUCCESS"

    def _prepTestDecoratorContext(self, testsToRun: List[str]) -> dict:
        """
        Prepares the decorator context
        """
        try:
            stack = _lytixTestContext.get().copy()
            stack["testsToRun"] = testsToRun
            stack["traceId"] = str(uuid.uuid4())
            stack["messages"] = []
            stack["modelName"] = None
            _lytixTestContext.set(stack)

            return {}
        except Exception as e:
            self._logger.error(f"Error setting up decorator context {e}")

    def _getTestContext(self):
        """
        Gets the current context
        """
        existingContext = _lytixTestContext.get()

        if existingContext["traceId"] is None:
            self._logger.debug(
                "No traceId in context, make sure you have setup the correct decorator"
            )
            return {}, False
        return existingContext, True

    def setMessage(self, message: str, role: str):
        """
        Sets model input in the current context
        @param role: One of 'user', 'system', 'assistant'
        """
        messageToAdd = {"role": role, "content": message}

        # Try to set on Test context
        contextTest, validTest = self._getTestContext()
        if validTest:
            contextTest["messages"].append(messageToAdd)
            _lytixTestContext.set(contextTest)


lytixTest = LytixTestWrapper()



import asyncio
from collections import defaultdict
import contextvars
from functools import wraps
import logging
import time
from typing import Any, Callable, DefaultDict, List, TypeVar, cast
import uuid

from lytix_py.MetricCollector.MetricCollector import MetricCollector

"""
@see https://mypy.readthedocs.io/en/stable/generics.html#declaring-decorators
"""
F = TypeVar("F", bound=Callable[..., Any])


_lytixTraceContext: contextvars.ContextVar[DefaultDict[str, str]] = (
    contextvars.ContextVar(
        "lytixTraceContext",
        default=defaultdict(
            lambda: {
                "traceId": None,
                "modelName": None,
                "messages": None,
                "userIdentifier": None,
                "sessionId": None,
                "backingModelId": None,
            },
        ),
    )
)


class LytixWrapper:
    _logger = logging.getLogger("lytix")

    def trace(
        self,
        *kwargs,
        modelName: str,
    ) -> Callable[[F], F]:
        """
        Main trace decorator.
        @param modelName: Model name we are tracing
        """

        def decorator(func: F) -> F:
            """
            First lets see if this is an async function
            """
            isAsyncFunction = asyncio.iscoroutinefunction(func)

            """
            Wrap our function in a context state
            """
            if isAsyncFunction:
                return self._asyncTraceWrapper(func, modelName)
            else:
                return self._syncTraceWrapper(func, modelName)

        return decorator

    def _syncTraceWrapper(self, func: F, modelName: str):
        """
        Sync wrapper for the trace decorator
        """

        @wraps(func)
        def syncDecorator(*args, **kwargs):
            traceContext = self._prepTraceDecoratorContext(modelName)
            startTime = time.time()
            result = func(*args, **kwargs)
            responseTime = int(
                (time.time() - startTime) * 1000
            )  # Convert to milliseconds

            """
            If we have model IO set, send it to HQ
            """
            self.reportModelIO(responseTime)

            return result

        return cast(F, syncDecorator)

    def _asyncTraceWrapper(self, func: F, modelName: str):
        """
        Async wrapper for the trace decorator
        """

        @wraps(func)
        async def asyncDecorator(*args, **kwargs):
            traceContext = self._prepTraceDecoratorContext(modelName)
            startTime = time.time()
            result = await func(*args, **kwargs)
            responseTime = int(
                (time.time() - startTime) * 1000
            )  # Convert to milliseconds

            """
            If we have model IO set, send it to HQ
            """
            self.reportModelIO(responseTime)

            return result

        return cast(F, asyncDecorator)

    def reportModelIO(self, responseTime: int):
        """
        Reports model input and output to HQ
        """

        context, valid = self._getTraceContext()
        if not valid:
            return

        messages = context["messages"]
        modelName = context["modelName"]

        if messages is None or modelName is None:
            self._logger.error(
                "Missing messages or modelName in context, make sure you have setup the correct decorator"
            )
            return

        """
        Pull optional fields now as well
        """
        userIdentifier = context["userIdentifier"]
        sessionId = context["sessionId"]
        backingModelId = context["backingModelId"]

        """
        Send to HQ
        """
        MetricCollector.captureModelIO(
            modelName=modelName,
            messages=messages,
            userIdentifier=userIdentifier,
            sessionId=sessionId,
            modelResponseTime=responseTime,
            backingModelId=backingModelId,
        )

    def _prepTraceDecoratorContext(self, modelName: str) -> dict:
        """
        Prepares the decorator context
        """
        try:
            stack = _lytixTraceContext.get().copy()
            stack["modelName"] = modelName
            stack["traceId"] = str(uuid.uuid4())
            stack["messages"] = []
            stack["userIdentifier"] = None
            stack["sessionId"] = None
            _lytixTraceContext.set(stack)
            return {}
        except Exception as e:
            self._logger.error(f"Error setting up trace decorator context {e}")

    def _getTraceContext(self):
        """
        Gets the current context
        """
        existingContext = _lytixTraceContext.get()
        if existingContext["traceId"] is None:
            self._logger.debug(
                "No traceId in context, make sure you have setup the correct decorator"
            )
            return {}, False
        return existingContext, True

    def setMessage(self, message: str, role: str):
        """
        Sets model input in the current context
        @param role: One of 'user', 'system', 'assistant'
        """
        messageToAdd = {"role": role, "content": message}

        #  Set on trace context
        contextTrace, validTrace = self._getTraceContext()
        if validTrace:
            contextTrace["messages"].append(messageToAdd)
            _lytixTraceContext.set(contextTrace)

    def setUserIdentifier(self, userIdentifier: str):
        """
        Sets user identifier in the current context
        """
        context, valid = self._getTraceContext()
        if valid:
            context["userIdentifier"] = userIdentifier
            _lytixTraceContext.set(context)

    def setSessionId(self, sessionId: str):
        """
        Sets session id in the current context
        """
        context, valid = self._getTraceContext()
        if valid:
            context["sessionId"] = sessionId
            _lytixTraceContext.set(context)

    def setBackingModel(self, backingModelId: str):
        """
        Sets backing model id in the current context
        (e.g. "GPT_3_5_TURBO" | "GPT_4" | "LLAMA_3_8B_INSTRUCT" | "LLAMA_3_70B_INSTRUCT" | "MISTRAL_7B_INSTRUCT" | "MISTRAL_8X7B_INSTRUCT")
        """
        context, valid = self._getTraceContext()
        if valid:
            context["backingModelId"] = backingModelId
            _lytixTraceContext.set(context)


lytix = LytixWrapper()



from langchain_core.documents import Document
from langchain_core.outputs.llm_result import LLMResult
from typing import Any, Dict
from langchain.callbacks.base import BaseCallbackHandler

from lytix_py.MetricCollector.MetricCollector import MetricCollector


class LytixLangchainCallback(BaseCallbackHandler):
    documentsInChain = []

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        """Run when LLM ends running."""
        try:
            lytixEventId = response.generations[0][0].message.response_metadata[
                "lytix-proxy-payload"
            ]["lytixEventId"]

            if lytixEventId is None:
                import warnings

                warnings.warn("No lytixEventId found")
                return

            if len(self.documentsInChain) > 0:
                for chunk in self.documentsInChain:
                    MetricCollector._captureRAGChunk(chunk.page_content, lytixEventId)
        except Exception as e:
            warnings.warn("Error capturing RAG chunk", e)

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> Any:
        """Run when chain ends running."""
        if isinstance(outputs, list):
            """
            Check to make sure every item in the list is of type Document
            """
            allDocuments = True
            for output in outputs:
                if not isinstance(output, Document):
                    allDocuments = False
            if allDocuments:
                self.documentsInChain.extend(outputs)



from datetime import datetime
import json
from lytix_py.LAsyncStore.LAsyncStore import LAsyncStore
from lytix_py.LLogger.LLogger import LLogger
from lytix_py.MetricCollector.MetricCollector import MetricCollector
import traceback


class LError(Exception):
    def __init__(self, message: str, errorMetadata: dict = {}):
        super().__init__(message)
        try:
            """
            First get all logs from our async store
            """
            logs = LAsyncStore.getLogs()

            """
            Add a new log with our error
            """
            error_trace = "".join(traceback.format_stack())
            logs.append(
                json.dumps(
                    {
                        "name": "LError",
                        "hostname": "",
                        "pid": -1,
                        "level": 50,
                        "msg": f"Error: {message}\n{error_trace}",
                        "time": datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
                    }
                )
            )

            """
            Send the logs+ metadata to lytix
            """
            MetricCollector._captureLError(
                errorMsg=message,
                errorMetadata=errorMetadata,
                logs=logs,
            )
        except Exception as e:
            print("Error sending Lytix metric", e)



from .MetricCollector.MetricCollector import MetricCollector
from .LLogger.LLogger import LLogger
from .envVars import LytixCreds
from .LError import LErrorIncrement, LError
from .Lytix.Lytix import lytix
from .Lytix.LytixTest import lytixTest
from .LytixLangchainCallback.LytixLangchainCallback import LytixLangchainCallback



import os


class LytixEnvVars:
    LX_BASE_URL = os.environ.get("LX_BASE_URL", "https://api.lytix.co")
    LX_API_KEY = os.environ.get("LX_API_KEY")

    def setAPIKey(self, apiKey):
        self.LX_API_KEY = apiKey

    def setBaseURL(self, baseURL):
        self.LX_BASE_URL = baseURL

    def validateEnvVars(self):
        if self.LX_BASE_URL is None:
            print("LX_BASE_URL envvar is not set")
        if self.LX_API_KEY is None:
            print(
                "LX_API_KEY envvar is not set. Make sure to set it via LytixCreds.setAPIKey before making requests"
            )


"""
Always validate when importing this file
@note This will not hard break, just print an error
"""
LytixCreds = LytixEnvVars()
LytixCreds.validateEnvVars()



import datetime
from typing import Callable, List
import uuid
import requests
from urllib.parse import urljoin
import os.path
import time

from lytix_py.LAsyncStore.LAsyncStore import LAsyncStoreClass
from lytix_py.LLogger.LLogger import LLogger
from lytix_py.envVars import LytixCreds

"""
 Main class to collect and report metrics
 back to HQ
"""


class _MetricCollector:
    processing_metric_mutex: int = 0

    def __init__(self):
        self.processing_metric_mutex = 0

    def _get_base_url(self):
        return urljoin(LytixCreds.LX_BASE_URL, "/v2/metrics")

    def _get_base_test_url(self):
        return urljoin(LytixCreds.LX_BASE_URL, "/v2/test")

    def _sendPostRequest(self, endpoint: str, body: dict, url: str = None):
        """
        Interal wrapper to send a post request
        """
        if url is None:
            url = self._get_base_url()
        urlToUse = os.path.join(url, endpoint)
        headers = {
            "Content-Type": "application/json",
            "lx-api-key": LytixCreds.LX_API_KEY,
        }
        response = requests.post(urlToUse, headers=headers, json=body)
        if response.status_code != 200:
            print(
                f"Failed to send post request to {urlToUse} with status code {response.status_code} and response {response.text}"
            )

        return response

    def _sendGetRequest(self, endpoint: str, url: str = None):
        """
        Interal wrapper to send a get request
        """
        if url is None:
            url = self._get_base_url()
        urlToUse = os.path.join(url, endpoint)
        headers = {
            "Content-Type": "application/json",
            "lx-api-key": LytixCreds.LX_API_KEY,
        }
        response = requests.get(urlToUse, headers=headers)
        if response.status_code != 200:
            raise Exception(
                f"Failed to send get request to {urlToUse} with status code {response.status_code} and response {response.text}"
            )
        return response

    """
    Increment a given metric
    """

    def increment(
        self, metricName: str, metricValue: int = 1, metricMetadata: dict = None
    ):
        if LytixCreds.LX_API_KEY is None:
            return

        if metricMetadata is None:
            metricMetadata = {}
        body = {
            "metricName": metricName,
            "metricValue": metricValue,
            "metricMetadata": metricMetadata,
        }
        self._sendPostRequest("increment", body)

    """
    Capture a model input/output
    """

    def captureModelIO(
        self,
        modelName: str,
        messages: List[dict],
        metricMetadata: dict = None,
        userIdentifier=None,
        sessionId=None,
        logs: list[str] = [],
        modelResponseTime: int = None,
        backingModelId: str = None,
    ):
        if LytixCreds.LX_API_KEY is None:
            return

        if metricMetadata is None:
            metricMetadata = {}
        body = {
            "modelName": modelName,
            "messages": messages,
            "metricMetadata": metricMetadata,
            "userIdentifier": str(userIdentifier),
            "sessionId": str(sessionId),
            "logs": logs,
            "modelResponseTime": modelResponseTime,
            "backingModelId": backingModelId,
        }
        self._sendPostRequest("modelIO", body)

    """
    Capture a model io event while also capturing the time to respond
    """

    async def captureModelTrace(
        self,
        modelName: str,
        modelInput: str,
        callback: Callable[[LLogger], str],
        metricMetadata: dict = {},
        userIdentifier=None,
        sessionId=None,
    ):
        asyncStore = LAsyncStoreClass()
        logger = LLogger(
            f"lytix-{modelName}-trace-{uuid.uuid4()}", asyncStore=asyncStore
        )
        if LytixCreds.LX_API_KEY is None:
            print("No Lytix API key found, skipping metric collection")
            return await callback(logger)

        startTime = time.time()
        modelOutput = await callback(logger)
        try:
            responseTime = int(
                (time.time() - startTime) * 1000
            )  # Convert to milliseconds
            # Capture modelIO event along with the response time and any logs
            logs = logger.asyncStore.getLogs()
            self.captureModelIO(
                modelName,
                modelInput,
                modelOutput,
                metricMetadata,
                userIdentifier,
                sessionId,
                logs,
            )
            self.increment(
                "model.responseTime",
                responseTime,
                {"modelName": modelName}.update(metricMetadata),
            )
        except Exception as err:
            self.logger.error(
                f"Failed to capture model trace: {err}", err, modelName, modelInput
            )
        finally:
            return modelOutput

    """
    Capture a metric trace event
    @note You likeley never need to call this directly
    """

    # @deprecated use _captureLError
    def _captureMetricTrace(
        self,
        metricName: str,
        metricValue: int,
        metricMetadata: dict = {},
        logs: list = [],
    ):
        self.processing_metric_mutex += 1

        body = {
            "metricName": metricName,
            "metricValue": metricValue,
            "metricMetadata": metricMetadata,
            "logs": logs,
        }

        self._sendPostRequest("increment", body)

        self.processing_metric_mutex -= 1

    """
    Capture LError event (logs + metadata)
    """

    def _captureLError(self, errorMsg: str, errorMetadata: dict = {}, logs: list = []):
        self.processing_metric_mutex += 1

        body = {
            "errorName": errorMsg,
            "timestamp": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
            "metadata": errorMetadata,
            "logs": logs,
        }

        self._sendPostRequest("lerror", body)

        self.processing_metric_mutex -= 1

    def _kickoffTestRun(
        self,
        messages: List[dict],
        testsToRun: List[str],
        traceId: str,
        testSuiteId: str,
        functionName: str,
    ):
        """
        Kickoff a test run in Lytix and returns the testRunId if present
        """
        self.processing_metric_mutex += 1

        body = {
            "messages": messages,
            "testsToRun": testsToRun,
            "traceId": traceId,
            "testSuiteId": testSuiteId,
            "functionName": functionName,
        }

        response = self._sendPostRequest("ioTest", body, url=self._get_base_test_url())
        if response.status_code != 200:
            print(
                f"🚨 Failed to kickoff test run with status code {response.status_code} and response {response.text}"
            )
            return False

        jsonResponse = response.json()

        if jsonResponse.get("valid", False) is False:
            print(
                f"🚨 Failed to kickoff test run with status code {response.status_code} and response {response.text}"
            )
            return False

        self.processing_metric_mutex -= 1
        return jsonResponse["testRunIds"]

    def _getTestStatus(self, testRunId: str):
        """
        Get the status of a test run
        """
        response = self._sendGetRequest(
            f"testRun/{testRunId}", url=self._get_base_test_url()
        )
        return response.json()

    def _finalizeTestGroup(self, testSuiteId: str):
        """
        Finalize a test group
        """
        response = self._sendPostRequest(
            f"finalizeTest", {"testSuiteId": testSuiteId}, url=self._get_base_test_url()
        )
        return response.json()

    def _createTestSuite(self):
        """
        Create a test suite
        """
        response = self._sendGetRequest(f"testSuite", url=self._get_base_test_url())
        result = response.json()
        return result["testSuiteId"]

    def _captureRAGChunk(self, chunkData: str, ioEventId: str):
        """
        Capture a RAG chunk
        """
        body = {
            "chunkData": chunkData,
            "ioEventId": ioEventId,
        }
        self._sendPostRequest("ragChunk", body)


MetricCollector = _MetricCollector()



import contextvars

from lytix_py.FastAPIMiddleware.MiddlewareState import getLytixContext

class LAsyncStoreClass:
    httpMiddleware = False

    def __init__(self):
        self.lytix_context_store = contextvars.ContextVar('lytix-context-store', default={'logs': []})
        self.httpMiddleware = False

    """
    Set our context store
    """
    def setHTTPMiddleware(self, httpMiddleware: bool):
        self.httpMiddleware = httpMiddleware
    
    """
    Append logs to our context store, always only storing 20 of the last logs so 
    we don't OOM 
    """
    def appendToLogs(self, log):
        if (self.httpMiddleware):
            getLytixContext()['logs'].append(log)
            getLytixContext()['logs'] = getLytixContext()['logs'][:20]
        else:
            self.lytix_context_store.get()['logs'].append(log)
            self.lytix_context_store.get()['logs'] = self.lytix_context_store.get()['logs'][:20]

    """
    Get logs from our async store
    """
    def getLogs(self):
        if (self.httpMiddleware):
            return getLytixContext()['logs']
        else:
            return self.lytix_context_store.get()['logs']

LAsyncStore = LAsyncStoreClass()





import ast, inspect


def StartTest(target):
    res = {}

    def visit_FunctionDef(node):
        res[node.name] = [ast.dump(e) for e in node.decorator_list]

    V = ast.NodeVisitor()
    V.visit_FunctionDef = visit_FunctionDef
    V.visit(compile(inspect.getsource(target), "?", "exec", ast.PyCF_ONLY_AST))
    return res



from starlette.types import ASGIApp, Receive, Scope, Send

from lytix_py.LAsyncStore.LAsyncStore import LAsyncStore
from lytix_py.FastAPIMiddleware.MiddlewareState import lytix_ctx_var

class LytixMiddleware:
    def __init__(
        self,
        app: ASGIApp,
    ) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ["http", "websocket"]:
            await self.app(scope, receive, send)
            return

        LAsyncStore.setHTTPMiddleware(True)

        lytixCtx = lytix_ctx_var.set({"logs": []})
        await self.app(scope, receive, send)
        lytix_ctx_var.reset(lytixCtx)



from contextvars import ContextVar

REQUEST_ID_CTX_KEY = "lytix-ctx-state"

lytix_ctx_var: ContextVar[str] = ContextVar(REQUEST_ID_CTX_KEY, default=None)

def getLytixContext() -> str:
    return lytix_ctx_var.get()
